/*
<H2>Licence Agreement</H2>
<PRE>InterfaceClassFinder package
Copyright (c) 2009 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.

Redistribution of the source of this package, with or without
modification, is NOT permitted.

Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.

2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.

3. Neither the name of the author nor Forward Computing and Control may be
   used to endorse or promote products derived from this software
   without specific prior written permission.

4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment:
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd.
    (http://www.forward.com.au)."


THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
</PRE>
 */
package au.com.forward.interfaceClassFinder;

import java.io.File;
import java.io.FileInputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

/**
 * This class scans the classpath and calls each consumer for each class found.
 * @author Matthew
 */
public class ClasspathScanner<T extends IClasspathProcessor> {

    public static final String DOT_CLASS = ".class";
    protected ConcurrentLinkedQueue<T> consumers = new ConcurrentLinkedQueue<T>();

    /**
     * Add a consumer to be called for each class found.
     * @param consumer the consumer to add
     */
    public void addConsumer(T consumer) {
        // no need to check if it is there as identityHashCode is the pseudo address of this consumer
        consumers.add(consumer);
    }

    /**
     * Remove a consumer.
     * @param consumer the consumer to remove (as determined by .equals() )
     * @return true if consumer found and removed.
     */
    public boolean removeConsumer(T consumer) {
        return consumers.remove(consumer);
    }

    /**
     * Return an array of all the current consumers registered with this scanner.<br>
     * Changes to this array does not add or remove consumers.
     * @return array of current consumers.
     */
    public T[] getConsumers() {
        return (T[]) consumers.toArray(new Object[0]);
    }

    /**
     * Scan the classpath and calls the registered consumers for each class found.
     * Quietly ignores files listed on the classpath that do not exist.
     */
    public void scanClassPath() {
        //get Classloader
        ClassLoader loader = Thread.currentThread().getContextClassLoader();
        // ClassLoader sysClassLoader = ClassLoader.getSystemClassLoader();
        if (!(loader instanceof URLClassLoader)) {
            throw new RuntimeException("Current Thread class loader not a URLClassLoader.");
        }
        //Get the URLs
        URL[] urls = ((URLClassLoader) loader).getURLs();
        for (int i = 0; i < urls.length; i++) {
            File f = new File(urls[i].getFile());
            // if file and exists and ends in .jar then open
            // if dir and exits then start root and look for .class files

            if (!(f.exists())) {
                //"Classpath entry '" + f.getPath() + "' does not exist.");
                continue; // try next one
            } //else 
            if (f.getName().endsWith(".jar")) {
                // process jar file
                processJarFile(f);
            } else {
                // process as root dir
                processFileDir(f, "", f.getPath());
            }
        }
    }

    /**
     * Scan the jar file for .class files to check
     * @param f the jar file to scan
     */
    protected void processJarFile(File f) {
        try {
            JarInputStream jarInputStream = new JarInputStream(new FileInputStream(f), false);
            JarEntry jarEntry = jarInputStream.getNextJarEntry();
            while (jarEntry != null) {
                if ((!jarEntry.isDirectory()) && jarEntry.getName().endsWith(DOT_CLASS)) {
                    String className = jarEntry.getName();
                    className = className.substring(0, className.lastIndexOf(DOT_CLASS));
                    className = className.replace('/', '.');
                    callConsumers(className, jarEntry, f);
                }
                jarEntry = jarInputStream.getNextJarEntry();
            }
        } catch (Throwable t) {
            if (!(t instanceof RuntimeException)) {
                throw new RuntimeException(t.toString(), t);
            } else {
                throw (RuntimeException) t;
            }
        }

    }

    /**
     * Process a file or directory. Recursively processes directories found.
     * @param f the file or directory to process
     * @param packagePath the current package name
     * @param rootDir the root directory
     * @param top true if this is the top of the package.
     */
    protected void processFileDir(File f, String packagePath, String rootDir) {
        if (f.isDirectory()) {
            // iterate files and dirs
            if (f.getPath().equals(rootDir)) {
                packagePath = "";
            } else {
                packagePath = packagePath.length() == 0 ? f.getName() : packagePath + "." + f.getName();
            }
            String[] files = f.list();
            for (String fileName : files) {
                File dirFile = new File(f, fileName);
                processFileDir(dirFile, packagePath, rootDir);
            }
        } else if (f.isFile() && f.getName().endsWith(DOT_CLASS)) {
            // only recognize lower class .class
            String className = f.getName();
            // strip .class
            className = className.substring(0, className.lastIndexOf(DOT_CLASS));
            className = packagePath.length() == 0 ? className : packagePath + "." + className;
            callConsumers(className, rootDir, f);
        } // else ignore
    }

    /**
     * Call all consumers for this class file.
     * @param className the class name
     * @param rootDir the root directory
     * @param classFile the class file
     */
    protected void callConsumers(String className, String rootDir, File classFile) {
        Iterator<T> it = consumers.iterator();
        while (it.hasNext()) {
            T consumer = it.next();
            try {
                consumer.processClass(className, rootDir, classFile);
            } catch (Throwable t) {
                System.err.println("Consumer " + consumer + " throw exception");
                t.printStackTrace();
            }
        }
    }

    /**
     * Call all consumers for this jar file entry.<br>
     * Note: the consumer must open the file and find the jarEntry passed in.
     * @param className the class name
     * @param jarEntry the jar entry
     * @param jarFile the jar file
     */
    protected void callConsumers(String className, JarEntry jarEntry, File jarFile) {
        Iterator<T> it = consumers.iterator();
        while (it.hasNext()) {
            T consumer = it.next();
            try {
                consumer.processClass(className, jarEntry, jarFile);
            } catch (Throwable t) {
                System.err.println("Consumer " + consumer + " throw exception");
                t.printStackTrace();
            }
        }
    }

    /**
     * Main class that list all the classes on the classpath and looks for duplicates.
     * @param args ignored
     */
    public static void main(String[] args) {
        final HashMap<String, String> map = new HashMap<String, String>();
        final List<String> dups = new ArrayList<String>();
        ClasspathScanner<IClasspathProcessor> scanner = new ClasspathScanner<IClasspathProcessor>();
        IClasspathProcessor consumer = new IClasspathProcessor() {

            public void processClass(String className, String rootDir, File classFile) {
                System.out.println("" + className + " in " + classFile.getPath());
                String location = map.get(className);
                if (location != null) {
                    // found dup
                    dups.add(""+className+" in "+classFile.getPath()+" duplicates class in "+location);
                } else {
                    // add it
                    map.put(className,classFile.getPath());
                }
            }

            public void processClass(String className, JarEntry jarEntry, File jarFile) {
                System.out.println("" + className + " in " + jarFile.getPath());
                String location = map.get(className);
                if (location != null) {
                    // found dup
                    dups.add(""+className+" in "+jarFile.getPath()+" duplicates class in "+location);
                } else {
                    // add it
                    map.put(className,jarFile.getPath());
                }
              }

        };
        scanner.addConsumer(consumer);
        scanner.scanClassPath();
        System.out.println();
        if (dups.isEmpty()) {
            System.out.println("No duplicate classes found.");
        } else {
            System.out.println("Found duplicate classes:-");
            System.out.println();
            for (String s:dups) {
                System.out.println(s);
                
            }
        }
    }
}
